#ifndef BTREE_H
#define BTREE_H

#include "../common/common.h"

/*
 * This function creates and writes to disk the original root
 * of the newly created B-Tree. Initially, the root will
 * contain zero keys and no rightmost child.
 */
void create_root();

/*
 * The function fetch_page reads a page from disk using the formula
 * page_num * PAGE_SIZE - PAGE_SIZE, e.g., if the page_num = 3 and
 * PAGE_SIZE = 1024, the B-Tree file will be accessed starting at
 * offset 2048, wherein the first page would be stored at offset 0,
 * and the second page at offset 1024.
 *
 * fetch_page installs the page in a dynamically allocated PageHdr
 * structure, which has a pointer to a linked-list of keys stored in 
 * dynamically allocated KeyRecord structures.
 */
PageHdr* fetch_page(long page_num);

/*
 * This function frees the space that was dynamimcally secured to hold a disk page.
 */
void free_page(PageHdr* page_ptr);

/*
 * This function writes a page to disk and then calls function
 * free_page to free the storage space that was allocated to the
 * dynamic data structures.
 *
 * The size of the page, PAGE_SIZE, is locked-in when the page is
 * written to disk. This is handled by the padding of character
 * zero's, e.g., if PAGE_SIZE = x and the number of bytes that are
 * stored in the page is num_bytes = n, there will be (x - n)
 * character byte zero's written to disk.
 */
void flush_page(PageHdr* page_ptr);

/*
 * Calculate the number of bytes and number of keys for a given page_ptr page image
 * and fill in the appropriate fields in the structure.
 */
void update_page_header(PageHdr* page_ptr);

/*
 * This function splits leaf and nonleaf pages. The page is split
 * according to the number of keys it contains. Half remain in the
 * current page (half + one in the case where the total number of
 * keys is an odd number), and half are transferred to a newly created
 * page. The two pages are written to disk (flush_page()). The
 * function returns the middle key which is to be inserted in the
 * parent page, in an "UpKey" structure.
 */
UpKey* split_page(PageHdr* page_ptr);

/*
 * Prior to execution of this function, the Root has already been
 * split by the function split_page(). The current Root contains
 * the left half of the Root that was split. This function creates
 * a new root (still located as the first page on disk) and installs
 * the key (up_key_ptr) that has propagated up to the Root.
 * The current root is installed as the left child of the new Root.
 */
void split_root(long root_page_num, UpKey* up_key_ptr);

/* Return page number of next page in the B-tree file. */
static inline long get_new_page_num() { return ffsize(fp_btree) / PAGE_SIZE + 1; }

/*
 * This function inserts a (key, offset) entry in the B-tree, even if it is empty initially.
 * If the B-Tree does exist, the key is sent to propagated_insert for insertion and any
 * possible page splits are handled there by recursively propagating down and then up the tree.
 * If propagated_insert returns the value NULL, this function terminates, otherwise
 * the root must be split (split_root()).
 */
void insert_in_tree(char* key, long doc_offset);

/*
 * This recursive function traverses through the B-Tree on a
 * directed path (beginning at the root and traversing down to
 * the leaf page that contains, or should contain the new key.
 * As pages are split, the middle keys are propagated up to the
 * parent pages for insertion. In the case where the root has
 * been split, the function returns the middle key which is to
 * be inserted in the new root, in the form of an "UpKey" structure.
 *
 * The function returns NULL otherwise.
 *
 * Note that the variable new_key points to a possible
 * propagated-up middle key and that key is sent to the function
 * InsertKey, which may, in turn, return a possible new
 * middle key which must be propagated-up to the parent page
 * for insertion. In the abstract, new_key travels down the
 * tree and middle_key travels upward.
 *
 * This function may be considered the core of the B-Tree
 * program.
 */
UpKey* propagated_insert(long page_num, char* key, long doc_offset);

/*
 * This recursive function searches a nonleaf page to find the
 * page number (child page) where the key should be stored. The
 * new key is compared with the keys within the B-Tree page. If
 * the new key is lexicograhically less than or equal to a stored
 * key, the page number of the child page is returned. If all
 * keys within the page have been searched and the new key is
 * lexicographically more than the last key, the page number of
 * the rightmost child is returned.
 */
long find_child_page(PageHdr* page_ptr, KeyRecord* key_list_iterator, char* key, int num_keys);

/*
 * This function inserts a key into a leaf page, after a posting has
 * been created. In the case where a key is already stored in the
 * B-Tree, the posting record is updated.
 *
 * If the page overflows, it splits it, flushes the two new pages and
 * return the value of the middle key in a structure "UpKey"
 * Otherwise, it just flushes back the old page and returns NULL
 */
UpKey* insert_in_leaf(PageHdr* page_ptr, char* key, long doc_offset);

/*
 * This function inserts a key into a nonleaf page. In the case
 * where the nonleaf page contains zero keys (the root), the first
 * key is inserted into the root and the only leaf page becomes
 * the rightmost child.
 *
 * It returns NULL, or a pointer to the key ("UpKey") to be propagated up.
 */
UpKey* insert_in_non_leaf(PageHdr* page_ptr, UpKey* middle_key);

/*
 * This recursive function searches a list of keys to find the
 * position where the new key should be inserted. The new key is
 * compared with the keys stored in the B-Tree until (1) a match is
 * found, (2) the new key is lexicographically less than the stored
 * key, (3) the last stored key is encountered and conditions (1)
 * or (2) have not been met. The function returns, for the above
 * cases: Case (1) -- the position of the matched key, Case (2) --
 * the position immediately preceding the current key being
 * compared, or Case (3) -- the position of the last key.
 *
 * Note that the variable count starts out with value zero and
 * increments with each recursive call.
 *
 * The variable key_list_iterator is a pointer that traverses
 * the list of keys.
 */
int find_insertion_pos(KeyRecord* key_list_iterator, char* key,
                       int* found, int num_keys, int count);

/*
 * This function compares the possible new key with a key stored
 * in the B-Tree. The function returns the value '0' if the keys
 * match, '1' if the new key is lexicographically less than the
 * stored key, and '2' if the new key is lexicographically more
 * than the stored key.
 */
int compare_keys(char* key, char* indexed_key);

/*
 * Recursively searches the subtree starting at the node stored at
 * the given page number for the given key.
 * Returns the offset from the postings file, or ERROR, if not found.
 */
long search_tree(long page_num, char* key);

/*
 * Searches the leaf at the node stored at the given page for the given key.
 * Returns the offset from the postings file, or NON_EXISTENT, if not found.
 */
long search_leaf(PageHdr* page_ptr, char* key);

#endif
