#include "btree.h"

/*
 * This recursive function traverses through the B-Tree on a
 * directed path (beginning at the root and traversing down to
 * the leaf page that contains, or should contain the new key.
 * As pages are split, the middle keys are propagated up to the
 * parent pages for insertion. In the case where the root has
 * been split, the function returns the middle key which is to
 * be inserted in the new root, in the form of an "UpKey" structure.
 *
 * The function returns NULL otherwise.
 *
 * Note that the variable new_key points to a possible
 * propagated-up middle key and that key is sent to the function
 * InsertKey, which may, in turn, return a possible new
 * middle key which must be propagated-up to the parent page
 * for insertion. In the abstract, new_key travels down the
 * tree and middle_key travels upward.
 *
 * This function may be considered the core of the B-Tree
 * program.
 */
UpKey* propagated_insert(long page_num, char* key, long doc_offset) {
    UpKey* middle_key;

    PageHdr* page_ptr = fetch_page(page_num);

    if (is_leaf(page_ptr)) {
        middle_key = insert_in_leaf(page_ptr, key, doc_offset);
    } else if ((!is_leaf(page_ptr))) {
        KeyRecord* key_list_iterator = page_ptr->key_list_ptr;  /* Pointer to list of keys */
        long child_page = find_child_page(page_ptr, key_list_iterator, key, page_ptr->num_keys);
        UpKey* new_key = propagated_insert(child_page, key, doc_offset);
        middle_key = insert_in_non_leaf(page_ptr, new_key);
    } else { /* impossible */
        printf("Prop-Insertion: Illegal PageType:%c", page_ptr->page_type);
        printf("in page# %d - aborting\n", (int) page_ptr->page_num);
        exit(-2);
    }

    return (middle_key);
}
