#include "btree.h"

/*
 * This function splits leaf and nonleaf pages. The page is split
 * according to the number of keys it contains. Half remain in the
 * current page (half + one in the case where the total number of
 * keys is an odd number), and half are transferred to a newly created
 * page. The two pages are written to disk (flush_page()). The
 * function returns the middle key which is to be inserted in the
 * parent page, in an "UpKey" structure.
 */
UpKey* split_page(PageHdr* page_ptr) {
    KeyRecord* pbefore;  /* points to the record before the middle */
    KeyRecord* pmiddle;  /* points to the middle-th record, which will be pushed upstairs */
    KeyRecord* pafter;   /* points to the successor of *pmiddle */


    /* Determine the number of keys to remain in current page and
    the number to be transferred to new page */
    int first_half_num_keys;
    if (page_ptr->num_keys % 2 == 0)
        first_half_num_keys = page_ptr->num_keys / 2;
    else
        first_half_num_keys = (page_ptr->num_keys / 2) + 1;

    /* Traverse list of keys up to the key before the middle */
    pbefore = page_ptr->key_list_ptr;
    for (int i = 1; i < (first_half_num_keys - 1); i++) {
        pbefore = pbefore->next;
    }
    pmiddle = pbefore->next;
    pafter  = pmiddle->next;

    /* Install the header of new page */
    PageHdr* new_page_ptr = (PageHdr*) malloc(sizeof(PageHdr));
    check_malloc(new_page_ptr, "new_page_ptr");
    new_page_ptr->page_type = page_ptr->page_type;
    new_page_ptr->page_num  = get_new_page_num();
    if (is_leaf(new_page_ptr))
        new_page_ptr->next_leaf_page_num   = page_ptr->next_leaf_page_num;
    if (!is_leaf(new_page_ptr))
        new_page_ptr->final_right_page_num = page_ptr->final_right_page_num;

    /* Transfer the keys of second half of page to new page */
    new_page_ptr->key_list_ptr = pafter;
    /* Adjust the last ptr of first half of page to be null */
    if (is_leaf(page_ptr))
        pmiddle->next = NULL;  /* the middle key stays on the leaf */
    else
        pbefore->next = NULL;  /* This way we have a proper B+ tree */

    /* Adjust the header of page being split */
    if (is_leaf(page_ptr))
        page_ptr->next_leaf_page_num   = new_page_ptr->page_num;
    if (!is_leaf(page_ptr))
        page_ptr->final_right_page_num = pmiddle->page_num;

    /* Create an "UpKey" node for the middle key, to be returned upwards */
    UpKey* upk = (UpKey*) malloc(sizeof(UpKey));
    check_malloc(upk, "upk");
    upk->left  = page_ptr->page_num;
    upk->right = new_page_ptr->page_num;
    upk->key   = strdup(pmiddle->stored_key);

    flush_page(page_ptr);
    flush_page(new_page_ptr);

    return upk;
}

