/*
 * This file contains routines that manage a hash table with separate chaining.
 * Much of the code is from the C book, p. 135
 */

#ifndef HASH_TABLE_H
#define HASH_TABLE_H

#include <stdlib.h>

#define HASH_SIZE 100

#define EOH -1  /* end of the hash table */

typedef struct nlist nlist;

struct nlist { /* basic table entry */
    char*  key;
    nlist* next;
};

static nlist* hash_table[HASH_SIZE];

/* hashing function */
static inline int hash(char* word) {
    int val;
    int i;

    for (val = 0, i = 0; word[i] != '\0'; i++) {
        val += word[i];
    }
    return (val % HASH_SIZE);
}

/* inserts a word in the hash table */
static inline void hash_table_insert(char* word) {
    nlist* np;
    int val;

    val = hash(word);
    np  = hash_table[val];
    while ((np != NULL)) {
        if (strcmp(np->key, word) == 0) { /* found */
            return;  /* do nothing */
        } else {
            np = np->next;
        }
    }

    /* if this line is reached, the word has not been found;
    insert it at the beginning of the chain */
    np = (nlist*) malloc(sizeof(nlist));
    check_malloc(np, "np in hash_table_insert()");

    np->key = malloc(strlen(word) + 1);
    strcpy(np->key, word);
    check_malloc(np->key, "np->key in hash_table_insert()");
    np->next = hash_table[val];
    hash_table[val] = np;
}

/* gives the next word in the hash table, returning EOH if that's the end */
static int position = -1;
static nlist* np = (nlist*) NULL;  /* np runs over all the nodes */

static inline int hash_table_get_next(char* word) {
    if (np != (nlist*) NULL) {
        strcpy(word, np->key);
        np = np->next;
        return TRUE;
    } else {
        position++;  /* examine next slot */
        /* skip empty slots */
        while ((hash_table[position] == NULL) && (position < HASH_SIZE)) {
            position++;
        }

        if (position == HASH_SIZE) {
            position = -1;  /* reset for the future */
            np       = NULL;
            return (EOH);
        } else {
            np = hash_table[position];
            strcpy(word, np->key);
            np = np->next;
            return TRUE;
        }
    }
}

static inline void hash_table_init() {
    for (int i = 0; i < HASH_SIZE; i++)
        hash_table[i] = NULL;
}

/* prints the hash table, for debugging */
static inline void hash_table_print() {
    for (int i = 0; i < HASH_SIZE; i++) {
        if ((np = hash_table[i]) != NULL) {
            printf("slot %d: ", i);
            for (; np != NULL; np = np->next) {
                printf("%s ", np->key);
            }
            printf("\n");
        }
    }
}

/* frees the space of the hash table */
static inline void hash_table_free() {
    for (int i = 0; i < HASH_SIZE; i++) {
        if ((np = hash_table[i]) != NULL) {
#ifdef DEBUG
            printf("freeing slot %d: ", i);
#endif

            while (np != NULL) {
#ifdef DEBUG
                printf(" %s ", np->key);
#endif

                free(np->key);

                nlist* npnext = np->next;
                free(np);
                np = npnext;
            }
            hash_table[i] = NULL;  /* reset it, for the future */
#ifdef DEBUG

            printf("\n");
#endif
        }
    }
}

#endif